/* File name     : vmelib.c
 * Creation date : September 1993
 * Author        : Y.Yasu, Online group, Physics department, KEK
 * Version       : 1.11
 * Modified date : August 1994, Y.Yasu
 *               : For HP-RT
 *               : Octover 1994, Y.Yasu
 *               : For fixing bug
 *               : November 1994, Y.Yasu
 *               : For fixing bug
 *               : December 1994, Y.Yasu
 *               : For implementation of VME Block Transfer
 *               : Octover 1995, Y.Yasu
 *               : For fixing bug
 *               : November 1995, Y.Yasu
 *               : For cleaning codes of the interrupt routines
 *		 : June 1997, Y.Yasu
 *		 : For Linux support on Bit3 VME adaptor(617,616)
 *		 : July 1997, Y. Yasu, K. Nakayoshi
 *		 : For Linux support on Bit3 VME adaptor(617,616)
 *		 : August 1999, K. Nakayoshi
 *               : For vmehb-1.4 
 */

#include	<stdio.h>
#include        <fcntl.h>
#include        <sys/mman.h>
#include	<sys/types.h>
#include	<sys/file.h>
#include        <sys/ioctl.h>
#include        <unistd.h>
#include        <sys/errno.h>
#include        "vmelib.h"
#define         SLOT 3

extern int      errno;

/* For VME devices */
static int vme_fd[SLOT];
static int vme_offset[SLOT];
static int vme_length[SLOT];

/* For Device File Path */
static int vme_dmapath;    /* file descriptor for DMA Transfer */

caddr_t vme_mapopen(int mode, caddr_t vmeaddr, size_t size)
{
    caddr_t vme_vaddr;	  /* virtual address */
    int slot;
    char vme_dev[15];
    int vme_mapsize;
    int vme_pagesize;

    vme_pagesize = getpagesize();

    switch (mode) {
    case VME_A16D32MODE :
    case VME_A16D16MODE :
#ifdef DEBUG
        printf("VME_A16MODE\n");
#endif
	slot = 0;
        if( size > VME_A16SIZE )
	    return (volatile caddr_t)-1;
	if( (vmeaddr < (volatile caddr_t)VME_A16BASE) || 
	    ((vmeaddr+size) > (caddr_t)(VME_A16BASE+VME_A16SIZE)) )
	    return (volatile caddr_t)-1;
        vme_offset[slot] = (int)vmeaddr&0xFFFF - VME_A16BASE;
	strcpy(vme_dev,"/dev/vme16d16");
	break;

    case VME_A24D16MODE :
    case VME_A24D32MODE :
#ifdef DEBUG
        printf("VME_A24MODE\n");
#endif
	slot = 1;
        if( size > VME_A24SIZE )
	    return (volatile caddr_t)-1;
	if( (vmeaddr < (volatile caddr_t)VME_A24BASE) || 
	    ((vmeaddr+size) > (volatile caddr_t)(VME_A24BASE+VME_A24SIZE)) )
	    return (volatile caddr_t)-1;
        vme_offset[slot] = (int)vmeaddr&0xFFFFFF - VME_A24BASE;
	strcpy(vme_dev,"/dev/vme24d16");
	break;

    case VME_A32D16MODE :
    case VME_A32D32MODE :
#ifdef DEBUG
        printf("VME_A32 MODE\n");
#endif
	slot = 2;
        if( size > VME_A32SIZE )
	    return (volatile caddr_t)-1;
/*	if( (vmeaddr < (volatile caddr_t)VME_A32BASE) || 
	    ((vmeaddr+size) > (volatile caddr_t)(VME_A32BASE+VME_A32SIZE)) )
	    return (volatile caddr_t)-1;
*/
        vme_offset[slot] = (int)vmeaddr - VME_A32BASE;
	strcpy(vme_dev,"/dev/vme32d32");
	break;
    };
    vme_mapsize = size + vme_offset[slot]%vme_pagesize;
    vme_length[slot] = vme_mapsize;
#ifdef DEBUG
    printf("vme_mapsize = %x\n", vme_length[slot]);
    printf("slot = %x\n", slot);
#endif
    if ((vme_fd[slot] = open(vme_dev, O_RDWR)) == -1)
        return (volatile caddr_t)-1;
    /* if mmap fails, return value becomes -1 */

    vme_vaddr = (caddr_t)mmap(0, vme_mapsize, PROT_READ | PROT_WRITE,
			      MAP_SHARED, vme_fd[slot],
			      vme_offset[slot]/vme_pagesize*vme_pagesize );

#ifdef DEBUG
    printf("vme_vaddr = %x\n", vme_vaddr);
    printf("vme_offset = %x\n", vme_offset[slot]);
    printf("vme_pagesize = %x\n", vme_pagesize);
#endif
    if (vme_vaddr == (volatile caddr_t) 0) {
        printf("errno = %d\n", errno);
        return (volatile caddr_t)-1;
    } else {
        return (vme_vaddr + (vme_offset[slot]%vme_pagesize));
    }
}

int vme_mapclose(int mode)
{
    int status, slot;

    switch (mode) {
    case VME_A16D32MODE :
    case VME_A16D16MODE :
        slot = 0;
	break;
    case VME_A24D32MODE :
    case VME_A24D16MODE :
        slot = 1;
	break;
    case VME_A32D32MODE :
    case VME_A32D16MODE :
        slot = 2;
	break;
    }

    status = munmap(NULL, (size_t)vme_length[slot]);

    if( status == -1 ) {
      printf("munmap error = %d\n", errno);
      return -1;
    }
    if( close(vme_fd[slot]) == -1) {
      printf("close error = %d\n", errno);
      return -1;
    }
    return 0;
}

/* DMA Transfer routines */

int vme_dmaopen(int mode, int dma_dev) {
  int slot;
  char vme_dev[15];
  
  switch (mode) {
  case VME_A24D32MODE :
  case VME_A24D16MODE :
    slot = 1;
    if(dma_dev == MEMORY) {
      strcpy(vme_dev,"/dev/vmm24d16");
    }
    if(dma_dev == FIFO) {
      strcpy(vme_dev,"/dev/vmf24d16");
    }
    break;
  case VME_A32D32MODE :
  case VME_A32D16MODE :
    slot = 2;
    if(dma_dev == MEMORY) {
      strcpy(vme_dev,"/dev/vmm32d32");
    }
    if(dma_dev == FIFO) {
      strcpy(vme_dev,"/dev/vmf32d32");
    }
    break;
  }

  if ((vme_fd[slot] = open(vme_dev, O_RDWR)) == -1) {
    printf("dmaopen Error\n");
    return -1;
  }
  vme_dmapath = vme_fd[slot];

  return 0;
}

int vme_dmaclose(int mode) {
  int slot;
  
    switch (mode) {
    case VME_A24D32MODE :
    case VME_A24D16MODE :
        slot = 1;
	break;
    case VME_A32D32MODE :
    case VME_A32D16MODE :
        slot = 2;
	break;
    }
    if( close(vme_fd[slot]) == -1) {
      printf("dmaclose Error\n");
      return -1;
    }
    return 0;
}

int vme_dmaread( int vmeaddr, void *buf, size_t count )
{
  int ret;
  
  if( ( lseek(vme_dmapath, vmeaddr, SEEK_SET) ) == -1 ){
    printf("lseek Error\n");      
    return errno;
  }
  if( ( ret = read( vme_dmapath, buf, count ) ) == -1 ){
    printf("Read Error\n");
    return errno;
  }
  return ret;
}

int vme_dmawrite( int vmeaddr, void *buf, size_t count )

{
  int ret;

  if( ( lseek(vme_dmapath, vmeaddr, SEEK_SET) ) == -1 ){
    printf("lseek Error\n");      
    return errno;
  }
  if( ( ret = write( vme_dmapath, buf, count ) ) == -1 ){
    printf("Write Error\n");
    return errno;
  }
  return ret;
}
